#include <GL/glew.h>
#include "mesh.h"
#include <QFile>
#include <QTextStream>
#include <QDebug>

void Mesh::loadSTL(const QString &filename)
{
	QFile file(filename);

	file.open(QIODevice::ReadOnly | QIODevice::Text);
	if (!file.isOpen())
		return;

	primitive_type = GL_TRIANGLES;

	QTextStream stream(&file);
	const QString &head = stream.readLine();
	if (head.left(6) == "solid ")	// ASCII format
	{
		name = head.right(head.size() - 6).toStdString();
		QString word;
		stream >> word;
		for(; word == "facet" ; stream >> word)
		{
			stream >> word;	// normal
			Vec3 n;
			stream >> n[0] >> n[1] >> n[2];
			n.normalize();

			stream >> word >> word;	// outer loop
			stream >> word;
			size_t start_idx = vertex.size();
			for(; word != "endloop" ; stream >> word)
			{
				Vec3 v;
				stream >> v[0] >> v[1] >> v[2];
				vertex.push_back(v);
			}
			for(size_t i = start_idx + 2 ; i < vertex.size() ; ++i)
			{
				index.push_back(start_idx);
				index.push_back(i - 1);
				index.push_back(i);
			}
			stream >> word;	// endfacet
		}
	}
	else			// BINARY format
	{
		file.setTextModeEnabled(false);
		file.seek(80);
		u_int32_t nb_triangles;
		file.read((char*)&nb_triangles, sizeof(nb_triangles));
		vertex.reserve(nb_triangles * 3U);
		normal.reserve(nb_triangles * 3U);
		index.reserve(nb_triangles * 3U);
		for(size_t i = 0 ; i < nb_triangles ; ++i)
		{
			Vec3f n, a, b, c;
#define READ_VECTOR(v)\
			do {\
				float f;\
				file.read((char*)&f,sizeof(float));	v[0] = f;\
				file.read((char*)&f,sizeof(float));	v[1] = f;\
				file.read((char*)&f,sizeof(float));	v[2] = f;\
			} while(false)

			READ_VECTOR(n);
			READ_VECTOR(a);
			READ_VECTOR(b);
			READ_VECTOR(c);

			index.push_back(vertex.size());
			index.push_back(vertex.size() + 1);
			index.push_back(vertex.size() + 2);
			vertex.push_back(a);
			vertex.push_back(b);
			vertex.push_back(c);

			u_int16_t attribute_byte_count;
			file.read((char*)&attribute_byte_count, sizeof(attribute_byte_count));
		}
	}
	recompute_normals();
}

void Mesh::saveSTL(const QString &filename) const
{
	QFile file(filename);

	file.open(QIODevice::WriteOnly);
	if (!file.isOpen())
		return;

	QByteArray header("QMeshViewer generated PLY");
	header.resize(80);
	file.write(header);

	u_int32_t nb_vtx(0), nb_triangles(0);
	std::vector<const Mesh*> qmesh;
	qmesh.push_back(this);
	while(!qmesh.empty())
	{
		const Mesh *cur = qmesh.back();
		qmesh.pop_back();
		for(size_t i = 0 ; i < cur->childs.size() ; ++i)
			qmesh.push_back(cur->childs[i]);

		nb_vtx += cur->vertex.size();
		switch(cur->primitive_type)
		{
		case GL_TRIANGLES:
			nb_triangles += cur->index.size() / 3;
			break;
		case GL_TRIANGLE_STRIP:
			nb_triangles += cur->index.empty() ? 0 : (cur->index.size() - 2);
			break;
		}
	}

	file.write((const char*)&nb_triangles, sizeof(nb_triangles));

#define WRITE_VECTOR(v)\
			do {\
				file.write((const char*)&v.x(),sizeof(float));\
				file.write((const char*)&v.y(),sizeof(float));\
				file.write((const char*)&v.z(),sizeof(float));\
			} while(false)

	qmesh.push_back(this);
	while(!qmesh.empty())
	{
		const Mesh *cur = qmesh.back();
		qmesh.pop_back();
		for(size_t i = 0 ; i < cur->childs.size() ; ++i)
			qmesh.push_back(cur->childs[i]);

		switch(cur->primitive_type)
		{
		case GL_TRIANGLES:
			for(size_t i = 0 ; i < cur->index.size() ; i += 3)
			{
				const Vec3f &a = cur->vertex[cur->index[i]];
				const Vec3f &b = cur->vertex[cur->index[i + 1]];
				const Vec3f &c = cur->vertex[cur->index[i + 2]];
				const Vec3f &n = (b - a) ^ (c - a);
				WRITE_VECTOR(n);
				WRITE_VECTOR(a);
				WRITE_VECTOR(b);
				WRITE_VECTOR(c);
				const u_int16_t attribute_byte_count(0);
				file.write((const char*)&attribute_byte_count, sizeof(attribute_byte_count));
			}
			break;
		case GL_TRIANGLE_STRIP:
			for(size_t i = 0 ; i + 2 < cur->index.size() ; ++i)
			{
				const Vec3f &a = cur->vertex[cur->index[i]];
				const Vec3f &b = cur->vertex[cur->index[i + 1]];
				const Vec3f &c = cur->vertex[cur->index[i + 2]];
				const Vec3f &n = (b - a) ^ (c - a);
				WRITE_VECTOR(n);
				WRITE_VECTOR(a);
				WRITE_VECTOR(b);
				WRITE_VECTOR(c);
				const u_int16_t attribute_byte_count(0);
				file.write((const char*)&attribute_byte_count, sizeof(attribute_byte_count));
			}
			break;
		}
	}
}
